package aspects.exceptionHandling;

import javax.servlet.http.*;
import javax.servlet.ServletException;

import java.io.PrintWriter;

import util.HTMLCode;

import org.aspectj.lang.SoftException;

public abstract aspect ServletsExceptionHandlingAspect 
        extends ExceptionHandlingAspect 
        percflow(ServletsExceptionHandlingAspect.servletsService()) {
     
	
	private boolean success = true;
	     
	/**
	 *  The PrintWriter to be used in the exception handling
	 */
    private PrintWriter printWriter;
    
	/**
	 * Stores the messages printed into the PrintWriter by the servlet, 
	 * which might be html forms or other content. 
	 * Those messages are stored here because instead of showing these
	 * content to the user if an exception is raised, the system should
	 * show the correspondent exception message. 
	 * If those messages were written into the PrintWriter, they could 
	 * not be rolled back. 
	 */      
	private StringBuffer buffer = new StringBuffer();

	/**
	 * Matches the servlet initialization
	 */
	pointcut servletInitialization(): execution(void HttpServlet+.init(..));
       
    /**
     * Matches the PrintWriter creation.
     */
    pointcut printWriterCreation(): target(HttpServletResponse) &&
                                    call(PrintWriter getWriter());
                                    
	/**
	 * Matches the PrintWriter's close method call.
	 */
	pointcut writerCloseExecution(): 
		this(HttpServlet+) && 
		call(* PrintWriter.close());

	/**
	 *  Matches the servlets services methods' execution
	 */
	public pointcut servletsService() : 
		within(HttpServlet+) && 
		(
			execution(void doPost(..)) 
		 	|| 
		 	execution(void doGet(..))
		);
	 
	void around() : servletInitialization() {
		try {
			proceed();
		} catch (ServletException ex) {
			System.out.println("*****************************************\n");
			System.out.println(" ERROR DURING SERVLET INITIALIZATION \n");
			System.out.println(" Error message: " + ex.getMessage() + "\n");
			System.out.println(" Location:      " + thisJoinPoint.getSourceLocation() + "\n");
			System.out.println("*****************************************");
		}
	}

    /**
     * Traping the PrintWriter creation to keep a reference
     * to it 
     */
    after() returning (PrintWriter pw): printWriterCreation() {
        printWriter = pw;
    }

    /**
     * Avoids executions of the writer's <tt>close<tt> method
     */
    void around(): writerCloseExecution() {
        // must do nothing!
        // close method is called by the exception handler
        // or by the advice after the normal termination of the 
        // servlet service
    }
    
    pointcut updateServletExecutions(String str): 
        this(HttpServlet+)   && 
        target(PrintWriter)  &&
        args(str)            &&
        call(* print*(String));
 
    /**
     * Intercepts writer's print messages
     */        
    void around(String str): updateServletExecutions(str) {
        buffer.append(str);
    }
   
    /**
     * Closes the writer
     */
    after(): servletsService() {
    	if (success) { //checks if the servlet's service was successfully executed
			printWriter.print(buffer.toString()); 
    	}
    	printWriter.flush();
        printWriter.close();
    }        

    /**
     * Method responsible for exceptions handling.
     * It shows a message to the user in a HTML page
     * using the servlet response's PrintWriter .
     */
    void exceptionHandling(Throwable exception) {

		this.success = false; //flags the unsuccessful servlet's service

		while (exception instanceof SoftException) {
			exception = ((SoftException) exception).getWrappedThrowable();
		}
		printWriter.println(HTMLCode.htmlPage("Error provided by the Exception Handling Aspect", 
		 								      exception.getMessage()));
		
		//ex.printStackTrace(printWriter); // prints the stack trace in the web browser
		//ex.printStackTrace();    // prints the stack trace in the console output
    }         

}